/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 27/08/22.
//

#include "grid.hpp"

#include "fggl/assets/loader.hpp"
#include "fggl/entity/gridworld/zone.hpp"

using namespace fggl::gfx::colours;

namespace demo {

	using namespace fggl::entity::grid;

	static void build_tileset(TileSet& tiles) {
		fggl::entity::grid::FloorTile empty{fggl::entity::grid::FloorTile::IMPOSSIBLE, BLACK};
		fggl::entity::grid::FloorTile ground{1, GREEN};
		tiles.m_floors.push_back(empty);
		tiles.m_floors.push_back(ground);

		fggl::entity::grid::WallTile noWall{};
		tiles.m_walls.push_back(noWall);

		fggl::entity::grid::WallTile solidWall{
			.render = true,
			.colour = DARK_SLATE_GRAY
		};
		tiles.m_walls.push_back(solidWall);
	}

	static void build_room(DemoGrid* area, fggl::math::vec2i center, fggl::math::vec2i size) {

		for (int yOffset = -size.y; yOffset <= size.y; ++yOffset) {
			auto yPos = yOffset + center.y;
			area->setWallAt(center.x - size.x, yPos, false, 1);
			area->setWallAt(center.x + size.x + 1, yPos, false, 1);
		}

		for (int xOffset = -size.x; xOffset <= size.x; ++xOffset) {
			auto xPos = center.x + xOffset;

			area->setWallAt(xPos, center.y - size.y, true, 1);
			area->setWallAt(xPos, center.y + size.y + 1, true, 1);

			for (int yOffset = -size.y; yOffset <= size.y; ++yOffset) {
				auto yPos = yOffset + center.y;
				area->setFloorAt(xPos, yPos, 1);
			}
		}
	}

	void build_doorway(DemoGrid* area, fggl::math::vec2i position, bool north, int size = 1) {
		for ( auto offset = 0; offset < size; offset++) {
			if ( north ) {
				area->setWallAt(position.x + offset, position.y, north, 0);
			} else {
				area->setWallAt(position.x, position.y + offset, north, 0);
			}
		}
	}

	static void build_test_env(DemoGrid* area) {
		area->clear();
		build_room(area, {5, 5}, {4,4});
		build_room(area, {11, 5}, {1,1});
		build_room(area, {17, 5}, {4,4});
		build_doorway(area, {10, 5}, false, 1);
		build_doorway(area, {13, 5}, false, 1);

		build_room(area, {25, 5}, {3,3});

		// player
		auto& manager = area->entities();
		{
			auto player = manager.create();
			auto& cellPos = manager.add<CellPos>(player);
			cellPos.pos = {5,5};
		}
	}

	GridScene::GridScene(fggl::App &app) : Game(app), m_tiles(), m_grid(nullptr) {
	}

	void GridScene::activate() {
		Game::activate();
		fggl::debug::log(fggl::debug::Level::info, "GridScene::activate()");


		// fake loading the tileset
		if ( m_tiles.m_floors.empty() ) {
			build_tileset(m_tiles);
			//auto* assetLoader = m_owner.service<fggl::assets::Loader>();
			//assetLoader->load("tileset_base.yml", ASSET_TILESET);
		}

		// create the grid world
		m_grid = std::make_unique<DemoGrid>(m_tiles);
		build_test_env(m_grid.get());
	}

	void GridScene::deactivate() {
		m_grid = nullptr;
	}


	constexpr float DRAW_SIZE = 64.0F;
	constexpr float DRAW_HALF = DRAW_SIZE / 2.0F;
	constexpr float WALL_HALF = 2.5F;

	static void render_grid(fggl::gfx::Paint& paint, fggl::entity::grid::Area2D<255, 255>& grid) {
		const fggl::math::vec2 wallOffsetNorth {0, -DRAW_HALF};
		const fggl::math::vec2 wallOffsetWest {-DRAW_HALF, 0};

		for (int i=0; i <= 31; ++i) {
			for (int j=0; j <= 31; ++j) {
				const fggl::math::vec2 drawPos{i * DRAW_SIZE, j*DRAW_SIZE};

				// draw floor
				{
					auto &floor = grid.floorAt(i, j);
					fggl::gfx::Path2D tileGfx = fggl::gfx::make_rect(drawPos, {DRAW_HALF, DRAW_HALF}, floor.colour);
					paint.fill(tileGfx);
				}

				auto& wallNorth = grid.wallAt(i, j, true);
				if ( wallNorth.render ) {
					fggl::gfx::Path2D tileGfx = fggl::gfx::make_rect(drawPos + wallOffsetNorth, {DRAW_HALF, WALL_HALF}, wallNorth.colour);
					paint.fill(tileGfx);
				}

				auto& wallWest = grid.wallAt(i, j, false);
				if ( wallWest.render ) {
					fggl::gfx::Path2D tileGfx = fggl::gfx::make_rect(drawPos + wallOffsetWest, {WALL_HALF, DRAW_HALF}, wallWest.colour);
					paint.fill(tileGfx);
				}
			}
		}
	}


	static void render_objects(fggl::gfx::Paint& paint, DemoGrid& grid) {
		auto& manager = grid.entities();

		auto entities = manager.find<CellPos>();
		for (const auto& entity : entities ) {
			//auto& sprite = manager.get<Sprite>(entity);
			auto& cellPos = manager.get<CellPos>(entity);

			// convert grid pos to world pos
			fggl::math::vec2f drawPos = cellPos.pos;
			drawPos *= DRAW_SIZE;

			auto shape = fggl::gfx::make_shape(drawPos, DRAW_HALF, 3);
			paint.fill(shape);
		}
	}

	//float progress = 0.0f;
	void GridScene::render(fggl::gfx::Graphics &gfx) {
		Game::render(gfx);

		fggl::gfx::Paint paint;
		render_grid(paint, *m_grid);
		render_objects(paint, *m_grid);

		/*
		// draw test shapes to check grid alignment
		for (int sides = 3; sides <= 25; ++sides) {
			auto hexTest = fggl::gfx::make_shape(fggl::math::vec2{sides, 5} * DRAW_SIZE, DRAW_HALF, sides, {1.0F - (sides / 25.0F), 0.0f, sides / 25.0f});
			paint.fill(hexTest);
		}

		// draw with edges
		for (int sides = 3; sides <= 25; ++sides) {
			auto hexTest = fggl::gfx::make_shape(fggl::math::vec2{sides, 6} * DRAW_SIZE, DRAW_HALF, sides, {1.0F - (sides / 25.0F), 0.0f, sides / 25.0f});
			paint.stroke(hexTest);
		}

		// draw test arcs
		for (int sides = 0; sides < 12; ++sides) {
			float endAngle = progress * (M_PI * 2);

			float startAngle = sides/12.0F * (M_PI * 2);
			float endAngle2 = endAngle + startAngle;

			auto hexTest = fggl::gfx::make_arc(fggl::math::vec2{sides + 3, 7} * DRAW_SIZE, DRAW_HALF, startAngle, endAngle2, fggl::gfx::colours::CYAN);
			paint.fill(hexTest);
		}

		// draw sweep
		for (int sides = 2; sides <= 25; ++sides){
			float angle = progress * (M_PI * 2);
			float sliceSize = 1 / (float)sides * (M_PI * 2);

			float startAngle = angle;
			float endAngle = startAngle + sliceSize;

			auto hexTest = fggl::gfx::make_arc(fggl::math::vec2{sides + 3, 8} * DRAW_SIZE,
											   DRAW_HALF,
											   startAngle,
											   endAngle,
											   fggl::gfx::colours::CYAN);
			paint.fill(hexTest);
		}


		progress += 0.01F;
		if ( progress > 1.0F) {
			progress = 0.0F;
		}*/

		gfx.draw2D(paint);
	}

} // namespace demo
